/*
**		@(#)MTG-Ethernet:glob.h	9.7	93/04/01
*/
#ident	"@(#)MTG-Ethernet:glob.h	9.7	93/04/01"


/*******************************************************************************
* include's
*******************************************************************************/
#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysmacros.h>
#include <sys/errno.h>		/* return errors for ioctl	*/
#include <sys/cmn_err.h>	/* cmn_err() and prt_where	*/
#include <local/mvme147.h>	/* declaration of struct pcc	*/
#include <local/kern.h>		/* defs for kernel entry points	*/
#include <local/mtg.h>		/* controller interface for ioctl */
#include <sys/signal.h>		/********************************/
#include <sys/fs/s5dir.h>	/*****  the user structure  *****/
#include <sys/user.h>		/********************************/
#include <sys/immu.h>		/********************************/
#include <sys/region.h>		/*****  PMMU declarations   *****/
#include <sys/proc.h>		/********************************/

#ifndef NPDEPP			/* Number of PDE's per page	*/
#define NPDEPP 1
#endif

#define uvtophys(uva)	(((caddr_t)pg_getaddr(uvtopde(uva)))+poff(uva))


/*******************************************************************************
* common constants
*******************************************************************************/
#define	RLEN	3	/* two's exponents of numbers of receive	*/ 
#define	TLEN	0	/* and transmit buffer descriptors (0-7)	*/
#if (TLEN != 0)
	ERROR	Must change the entire transmit logic !
#endif
#define	ERR_SIZE	20	/* size of error messages chain  (cm_t)	*/

#define	LANVECT		0x44	/* LANCE interrupt vector		*/
#define	RBUFLEN		/* length of Ethernet  receive buffer		*/\
			(MAXRBUF - (ETH_MAX_TU-TG_MAXDPL))
#define	INVALID_TGT	((tgt_t *)-1)	/* pointer to no tgt		*/

#define PCC_tpr		V147_PCC->tpr2	/* Tick Timer 2 preload reg.	*/
#define	PCC_tcr		V147_PCC->tcr2	/* Tick Timer 2 control reg.	*/
#define PCC_ticr	V147_PCC->ticr2	/* Tick Timer 2 int & contr reg */
#define PCC_licr	V147_PCC->licr	/* LANCE int & contr reg.	*/
#define	LNC_rdp		V147_LNC->le_rdp /* LANCE Register Data Port	*/
#define	LNC_rap		V147_LNC->le_rap /* LANCE Register Address Port	*/
#define	ETH_addr	V147_NVRAM->nv_net.net_esd.station_addr
#define	ETH_mask	V147_NVRAM->nv_net.net_esd.group_mask


/*******************************************************************************
* fatal error messages
*******************************************************************************/
#define	TF_CNVR	"MTG driver:\
 BBram has been corrupted. Ethernet station address must be reset"
#define	TF_ERRE	"MTG driver: ERR_SIZE must be increased"
#define	TF_CMPE	"MTG driver: CMP_SIZE must be increased"
#define	TF_RCVB	"MTG driver:\
 receive buffer allocated above 0x01000000 or not quadword aligned"
#define	TF_LNCI	"MTG driver: LANCE error during initialization. csr0 = 0x%x"
#define	TF_LNCA	"MTG driver:\
 LANCE initialisation block above 0x01000000 or not word aligned"
#define	TF_SCCF	"MTG driver: scan() consistency check failed"
#define	TF_LNCL	"MTG driver: LANCE error in mtg_lintr(). csr0 = 0x%x"
#define	TF_OFLO	"MTG driver: overflow error.  Halting MTG."
#define	TF_BUFF	"MTG driver: buffer error.  Halting MTG."
#define	TF_CMPL "MTG driver: CMP_SIZE must be increased"
#define	TF_DPPL	"MTG driver: DPP_SIZE must be increased"
#define	TF_UFLO	"MTG_driver: transmit underflow error"
#define	TF_BUFE	"MTG driver: transmit buffer error"
#define	TF_LICF	"MTG driver:\
 mtg_lintr() consistency check failed. semaphore = %d"
#define	TF_UNRX "MTG driver: unespected Rx interrupt. currmdndx = %d"
#define	TF_UNTX "MTG driver: unespected Tx interrupt"


/*******************************************************************************
* common structures and related constants
*******************************************************************************/
/*
 * TG extended timetable entry
 */
typedef struct {	/* the two first members must be same as tt_e	*/
	ushort	ete_Dt;		/* base ticks from prev. entry to this	*/
	ushort	ete_len;	/* base length of this packet in bytes	*/
	ushort	ete_rDt;	/* real ticks from prev. entry to this	*/
	ushort	ete_rlen;	/* real length of this packet in bytes	*/
} ett_e;
#define	ETTE_SZ	sizeof(ett_e)

/*
 * TGT or ctl message descriptor
 */
typedef union tgt {
	struct {			/* CTL section			*/
		ushort	type;		/* type == CTL_D		*/
		ushort	active;		/* activity (true or false)	*/
		union tgt	*nxt;	/* forward chain pointer	*/
		int	counter;	/* tick counter (0 or 1)	*/
		cm_t	ctl;		/* ctl message section		*/
	} ctl;
	struct {			/* TGT section			*/
		ushort	type;		/* type == TGT_D		*/
		ushort	active;		/* activity (true or false)	*/
		union tgt	*nxt;	/* forward chain pointer	*/
		int	counter;	/* tick counter			*/
		dp_t	*dp;	/* pointer to TG data packet		*/
		ett_e	*ett;	/* pointer to TG etimetable		*/
		ett_e	*cur;	/* pointer to current etimetable entry	*/
		uint	ettsz;	/* # of timetable entries		*/
	} tgt;
} tgt_t;
#define	d_type		ctl.type
#define	d_active	ctl.active
#define	d_nxt		ctl.nxt
#define	d_counter	ctl.counter
#define	TGT_SZ		sizeof(tgt_t)
/* tgt_t types							*/
#define	TGT_D	(ushort)'D'	/* TGT 				*/
#define	CTL_D	(ushort)'C'	/* ctl message descriptor	*/

/*
 * error message
 */
typedef	struct ebuf {
	struct ebuf	*err_nxt;	/* forward pointer		*/
	char		*err_uaddr;	/* controller address of err_msg */
	caddr_t		*err_back;	/* back pointer contains	 */
		/* physical address of controller's pointer to err_msg */
	cm_t		err_msg;	/* error message data		*/
} ebuf_t;

/*
 * receive buffer
 */
typedef	struct rbuf {
	struct rbuf	*r_nxt;		/* forward pointer		*/
	struct rbuf	*r_rlennxt;	/* points 1<<RLEN rbuf's forward*/
	char		*r_uaddr;	/* controller address of r_.data*/
	caddr_t		*r_back;	/* back pointer contains	*/
		/* physical address of controller's pointer to r_.data */
	union {
		char	data[RBUFLEN];	/* generic data, i.e.	*/
		cm_t	cm_data;	/* control message ...	*/
		dp_t	dp_data;	/* or data packet	*/
	} r_;		/* receive buffer data		*/
} rbuf_t;


/*******************************************************************************
* common variables
*******************************************************************************/
extern char	mtg_ok;		/* false -> fatal hard or soft error	*/
extern char	mtg_in_use;	/* device is open(2)ed			*/
extern char	mtg_inited;	/* switched by TG_INIT and TG_HALT	*/
extern int	(*mtg_oldintvec)();	/* old int vector to be restored*/

extern struct eth_statistics {
	int	st_tot;		/* total number of packets	*/
	int	st_def;		/* deferred packets count	*/
	int	st_one;		/* exactly one retry count	*/
	int	st_more;	/* more than one retry count	*/
	int	st_rtry;	/* retry errors			*/
}	mtg_stat;		/* structure for Ethernet tx statistics	*/

extern caddr_t	mtg_crbuf;	/* address of contiguous receive buffer	*/
extern caddr_t	mtg_vcrbuf;	/* kernel virtual address of crbuf	*/
extern uint	mtg_crbsize;	/* size (in clicks) of crbuf		*/
extern caddr_t	mtg_lanrbuf;	/* start of space for LANCE structures	*/
extern cmp_e	*mtg_ucmp_a;	/* controller address of cmp array	*/
extern dpp_e	*mtg_udpp_a;	/* controller address of dpp array	*/

extern time_t	wait_for_tick;	/* uticks to elapse until mtg_tick is called */
extern time_t	utick_del[16];	/* microtick delays			*/
extern time_t	*mtg_utick;	/* addr.of microtick (MTG system clock)	*/

extern struct le_init_block  le_iblk;	/* LANCE initialization block	*/

extern int	*mtg_sem;	/* address of TGT linked list semaphore	*/
extern uint	mtg_maxsem;	/* max semaphore without error message	*/
extern tgt_t	*mtg_head;	/* pointer to head of TGT linked list	*/
extern tgt_t	*mtg_scanindex;	/* TGT linked list scanning index	*/


/*******************************************************************************
* function declarations
*******************************************************************************/
void swab();		/* from standard C library		*/
void ma_int();		/* M68K LANCE interrupt hook		*/
void mtg_lint();	/* MTG LANCE interrupt hook		*/
void mtg_tint();	/* MTG Tick timer interrupt hook	*/
void mtg_ltryint();	/* MTG LANCE test interrupt hook	*/
char mtg_lantry();	/****************************************/
char mtg_tinit();	/*					*/
char mtg_linit();	/*					*/
void mtg_lintr();	/*	functions in low.c		*/
void mtg_scan();	/*					*/
void mtg_tstop();	/*					*/
void mtg_lstop();	/****************************************/
void mtg_halt();	/******	functions in high.c *************/


/*******************************************************************************
* lint and debug stuff
*******************************************************************************/
extern int	mtg__cmaj;
#define	DB_MAJ	mtg__cmaj
#define	DB_MIN	0
#include <local/kerndebug.h>

#ifdef lint
struct sndd	{int lint_dummy;};	struct response	{int lint_dummy;};
struct msgb	{int lint_dummy;};	struct proc	{int lint_dummy;};
struct inode	{int lint_dummy;};	struct file	{int lint_dummy;};
#endif	/* lint	*/
