/*
**		@(#)MTG-Ethernet:high.c	9.5	93/02/02
*/
#ident	"@(#)MTG-Ethernet:high.c	9.5	93/02/02"

#include "glob.h"		/* global MTG extern declar.	*/
#include <sys/buf.h>		/* included for B_READ #define	*/
#include <sys/map.h>		/* for malloc() and mfree()	*/


/* Size of static (maps) and dinamic (pool) structures	*/
#define	TGTPOOL_SZ	500		/* size of tgt pool		*/
#define	TGTMAP_SZ	(TGTPOOL_SZ + 2)/* size of tgt  map		*/
#define	ETTMAP_SZ	TGTMAP_SZ	/* size of etimetables map	*/

/* structures and functions for TGT allocation map	*/
static tgt_t			*mtg_tgtpool;
static struct map		mtg_tgtmap[TGTMAP_SZ];
#define	gettgt()		(tgt_t *)malloc(mtg_tgtmap, TGT_SZ)
#define	freetgt(tgt)		mfree(mtg_tgtmap, TGT_SZ, (uint)(tgt))
#define	inittgtmap(tgtpool_sz)    { mapinit(mtg_tgtmap, TGTMAP_SZ);	\
	((struct map *)mapstart(mtg_tgtmap))->m_size = 0;		\
	((struct map *)mapstart(mtg_tgtmap))->m_addr = 0;		\
	mfree(mtg_tgtmap, tgtpool_sz * TGT_SZ, (uint)mtg_tgtpool); }

/* structures and functions for etimetables allocation map	*/
static ett_e			*mtg_ettpool;
static struct map		mtg_ettmap[ETTMAP_SZ];
#define	getett(sz)		(ett_e *)malloc(mtg_ettmap, (sz) * ETTE_SZ)
#define	freeett(sz, ett)	mfree(mtg_ettmap, (sz) * ETTE_SZ, (uint)(ett))
#define	initettmap(ettpool_sz)    { mapinit(mtg_ettmap, ETTMAP_SZ);	\
	((struct map *)mapstart(mtg_ettmap))->m_size = 0;		\
	((struct map *)mapstart(mtg_ettmap))->m_addr = 0;		\
	mfree(mtg_ettmap, ettpool_sz * ETTE_SZ, (uint)mtg_ettpool); }



/*
 * Gracefully stops all MTG activity (undoes what TG_INIT did)
 */
void
mtg_halt()
{
	if (mtg_inited) {
		/* stop interrupt handlers activity	*/
		mtg_lstop();
		mtg_tstop();
		mtg_scanindex = INVALID_TGT;

		/* restore old interrupt vector and reinitialize LANCE */
		M68Kvec[LANVECT] = mtg_oldintvec;
		/* ma_init(); */

		/* unlock controller's cmp and dpp arrays */
		undma(mtg_ucmp_a, CMP_SIZE * sizeof(cmp_e), B_READ | B_WRITE);
		undma(mtg_udpp_a, DPP_SIZE * sizeof(dpp_e), B_READ | B_WRITE);

		mtg_inited = 0;
	}
} /* mtg_halt */



/*
 *   Gets 'size' contiguous pages within the first 16Mbytes and returns the
 * physical address of the first one. The pages are mapped in sysreg as
 * cache inhibited, and the kernel virtual address is returned in *vadr.
 */
caddr_t
getcrbpgs(size, nosleep, vadr)
register uint	size;
int		nosleep;
caddr_t		*vadr;
{
    caddr_t		padr;
    register pde_t	*pt;

    sync(); sync();

    /* return if size null or can't allocate size contiguous pages	*/
    if (size == 0 || (padr = getcpages(size, nosleep)) == NULL)
	return NULL;

    /* return if pages allocated are not within the first 16 Mbytes	*/
    if ((btoc(padr) + size) > btoc(0x01000000 /*16Mbytes*/)) {
	freecpages(padr, size);
	return NULL;
    }

    /* return if can't map the pages in kernel virtual memory	*/
    if ((*vadr = (caddr_t) sptalloc(size, 1, btoc(padr))) == NULL) {
	freecpages(padr, size);
	return NULL;
    }

    /* inhibit cacheing when using kernel virtual address	*/
    for (pt = kvtopde(*vadr); --size; pt += 1)
	pg_setci(pt);

    /* return physical address	*/
    return padr;

} /* getcrbpgs */



/*
 *   Unmaps the contiguous, cache inhibited pages allocated with getcrbpgs
 * from kernel memory map and releases them.
 * padr and vadr are the physical and kernel virtual addresses
 * of the first page to be freed.
 */
void
freecrbpgs(padr, vadr, size)
caddr_t	padr, vadr;
uint	size;
{
    if (size == 0)
	return;
    
    /* unmaps the pages from sysreg	*/
    sptfree(vadr, size, 0);

    /* deallocates the contiguous pages	*/
    freecpages(padr, size);

} /* freecrbpgs */



void
mtg_init()
{
#	if (!defined(DEBUG))
		/* if called by controller, free rbuf	*/
		if (mtg_crbuf) {
			freecrbpgs(mtg_crbuf, mtg_vcrbuf, mtg_crbsize);
			mtg_crbuf = NULL;
			mtg_crbsize = 0;
		}
#	endif

	mtg_ok = 1;
	mtg_in_use = 0;
	mtg_inited = 0;
} /* mtg_init */



/* ARGSUSED */
void
mtg_open(dev, mode)
dev_t	dev;
int	mode;
{
	if (!suser())
		return;
	if (mtg_in_use)
		{ u.u_error = EACCES; return; }
	if (!mtg_ok && eminor(dev) != TEST_MIN)
		{ u.u_error = EIO; return; }
	if (eminor(dev) != DEFAULT_MIN && eminor(dev) != TEST_MIN)
		{ u.u_error = ENODEV; return; }
	mtg_in_use = 1;
} /* mtg_open */




/* ARGSUSED */
void
mtg_close(dev, mode)
dev_t	dev;
int	mode;
{
	mtg_halt();
	mtg_in_use = 0;
} /* mtg_close */



/*
 *   Removes the dead control message descriptors from the TGT linked list.
 *   Called from most normal controller interaction commands.
 */
static void
clearcm()
{
	register tgt_t	*d, **dp;

	dp = &mtg_head;

	/* skip all TGT_D's: dp points to the pointer to the first CTL_D */
	while ((*dp) && ((*dp)->d_type ==  TGT_D))
		dp = &(*dp)->d_nxt;

	/* now check all CTL_D's and discard the dead ones	*/
	while (*dp) {
		if ((*dp)->d_active)
			dp = &(*dp)->d_nxt;
		else {
			d = (*dp);
			(*dp) = d->d_nxt;
			while (mtg_scanindex == d)
				;
			freetgt(d);
		}
	}
} /* clearcm */



/* ARGSUSED */
void
mtg_ioctl(dev, cmd, arg, mode)
dev_t	dev;
int	cmd, mode;
caddr_t	arg;
{
    mtg_ctl		ctl;
    register tgt_t	*d, **dp;
    register ett_e	*e;

    if (!mtg_ok && eminor(dev) != TEST_MIN)
	{ u.u_error = EIO; return; }
    switch (cmd) {
#define	COPYIN(f,t,c)	{ if(copyin(f, t, c)) {u.u_error = EFAULT; break; } }
#define	COPYOUT(f,t,c)	{ if(copyout(f, t, c)) {u.u_error = EFAULT; break; } }
/***************************************************************
* commands for crbuf allocation, deallocation and inquiring
***************************************************************/

    case TG_GETRBUF:
	mtg_inited = 0;
	COPYIN(arg, &ctl, sizeof(ctl.crb))
	if (mtg_crbuf != NULL || ctl.crb.physaddr != NULL)
	    { u.u_error = EINVAL; break; }
	if ((ctl.crb.physaddr = getcrbpgs(ctl.crb.size, 0, &mtg_vcrbuf)) ==NULL)
	    { u.u_error = ENOMEM; break; }
	COPYOUT(&ctl, arg, sizeof(ctl.crb))
	mtg_crbuf = ctl.crb.physaddr;
	mtg_crbsize = ctl.crb.size;
	break;

    case TG_FREERBUF:
	if (mtg_crbuf == NULL)
	    { u.u_error = EINVAL; break; }
	freecrbpgs(mtg_crbuf, mtg_vcrbuf, mtg_crbsize);
	mtg_crbuf = NULL;
	mtg_crbsize = 0;
	mtg_halt();
	break;

    case TG_ASKRBUF:
	ctl.crb.physaddr = mtg_crbuf;
	ctl.crb.size = mtg_crbsize;
	COPYOUT(&ctl, arg, sizeof(ctl.crb))
	break;
/***************************************************************
* commands for testing (only authorized personnel)
***************************************************************/

    case TG_RESET:
	if (eminor(dev) != TEST_MIN)
	    { u.u_error = EINVAL; break; }
	mtg_init();
	break;

/***************************************************************
* commands for normal controller interaction
***************************************************************/

    case TG_INIT:
	mtg_inited = 0;
	if (mtg_crbuf == NULL)
	    { u.u_error = EINVAL; break; }
	COPYIN(arg, &ctl, sizeof(ctl.init))

	/* stole LAN interrupt vector	*/
	mtg_oldintvec = M68Kvec[LANVECT];
	M68Kvec[LANVECT] = (int(*)())mtg_lint;

	/* initialise nvram Ethernet address	*/
	if ((V147_NVRAM->nv_bug.bug_eadr[0] & 0xf0) != 0x20) {
	    M68Kvec[LANVECT] = mtg_oldintvec;
	    (void) cmn_err(CE_WARN, TF_CNVR);
	    mtg_ok = 0;
	    return;
	}
	ETH_addr[0] = 0x08;
	ETH_addr[1] = 0x00;
	ETH_addr[2] = 0x3e;
	(void) bcopy((caddr_t)V147_NVRAM->nv_bug.bug_eadr,
		     (caddr_t)&ETH_addr[3], 3);

	/* check the ucrbuf address	*/
	if (!useracc(ctl.init.ucrbuf, ctob(mtg_crbsize), B_WRITE)) {
	    M68Kvec[LANVECT] = mtg_oldintvec;
	    u.u_error = EFAULT; break;}

	/* lock controller's cmp and dpp arrays in memory	*/
	mtg_ucmp_a = ctl.init.ucmp_a;
	mtg_udpp_a = ctl.init.udpp_a;
	if (!userdma(mtg_ucmp_a, CMP_SIZE * sizeof(cmp_e), B_READ | B_WRITE)) {
	    M68Kvec[LANVECT] = mtg_oldintvec;
	    u.u_error = EFAULT; break;}
	if (!userdma(mtg_udpp_a, DPP_SIZE * sizeof(dpp_e), B_READ | B_WRITE)) {
	    M68Kvec[LANVECT] = mtg_oldintvec;
	    undma(mtg_ucmp_a, CMP_SIZE * sizeof(cmp_e), B_READ | B_WRITE);
	    u.u_error = EFAULT; break; }

	/* this flag is tested by mtg_halt(), so we set it now	*/
	mtg_inited = 1;

	/* check the ett_pool minimum size	*/
	if (ctl.init.ettpool_sz < ETTPOOL_MIN_SZ)
	    { mtg_halt(); u.u_error = EINVAL; break; }

	/* zero down the entire crbuf */
	(void) bzero(mtg_crbuf, ctob(mtg_crbsize));

	/* initialise various pointers	*/
	mtg_utick = (time_t *)
		mtg_crbuf;
	mtg_sem = (int *)
		((int)mtg_utick + sizeof(mtg_utick));
	mtg_tgtpool = (tgt_t *)
		((int)mtg_sem + sizeof(mtg_sem));
	mtg_ettpool = (ett_e *)
		((int)mtg_tgtpool + TGTPOOL_SZ * TGT_SZ);
	mtg_ettpool = (ett_e *)
		(((uint)mtg_ettpool + 3) & ~3);
	mtg_lanrbuf = (caddr_t)
		((int)mtg_ettpool + ctl.init.ettpool_sz * sizeof(ett_e));
	/* mtg_lanrbuf MUST be longword aligned	*/
	mtg_lanrbuf = (caddr_t)(((uint)mtg_lanrbuf + 3) & ~3);

	/* check that the pools leave some space for rbuf's	*/
	if ((mtg_crbuf + ctob(mtg_crbsize)) <
	    (mtg_lanrbuf + 0x1000 +
	    ERR_SIZE * sizeof(ebuf_t) + (1<<RLEN) * sizeof(rbuf_t)))
		{mtg_halt(); u.u_error = ENOSPC ; break; }

	/* initialise the tgt map and the timetables map	*/
	inittgtmap(TGTPOOL_SZ);
	initettmap(ctl.init.ettpool_sz);

	/* initialise the TGT linked list variables	*/
	*mtg_sem = -1;
	mtg_maxsem = ctl.init.max_sem;
	mtg_head = NULL;
	mtg_scanindex = INVALID_TGT;

	/* hard and soft initialisation of devices	*/
	if ((u.u_error = mtg_tinit(&ctl.init.tick_len)) ||
	    (u.u_error = mtg_linit(ctl.init.ucrbuf)))
		{ mtg_halt(); break; }

	/* give the controller his parameters
	   (be careful to read his ones before!) */
	ctl.init.utick= (time_t *)((int)mtg_utick + ctl.init.ucrbuf-mtg_crbuf);
	ctl.init.sema = (int *)((int)mtg_sem + ctl.init.ucrbuf - mtg_crbuf);
	if (copyout(&ctl, arg, sizeof(ctl.init)))
	    { mtg_halt(); u.u_error = EFAULT; break; }

	break;

    case TG_HALT:
	mtg_halt();
	break;

    case TG_CTL:
    case TG_CREATE:
	if (!mtg_inited) {
	    scrivi("create:driver not inited\n");
	    u.u_error = EINVAL; goto create_err;
	}
	COPYIN(arg, &ctl, (cmd == TG_CTL) ? sizeof(ctl.ucm):sizeof(ctl.create))
	clearcm();

	/* get and fill a tgt_t	*/
	if ((d = gettgt()) == NULL)
		{ u.u_error = ENOSPC; goto create_err; }
	d->d_type = (cmd == TG_CTL) ? CTL_D : TGT_D;
	d->d_active = 1;
	d->d_counter = 1;

	if (cmd == TG_CTL) {	/* creation of a control msg descriptor	*/

	    /* copy control msg and check type	*/
	    if (copyin(ctl.ucm, &d->ctl.ctl, sizeof(cm_t)))
		{ u.u_error = EFAULT; goto create_err; }
	    if (!(d->ctl.ctl.cm_FODA_h.cm_code & CTL_CODE)) {
		scrivi2("ctl: wrong code 0x%x\n",
		    d->ctl.ctl.cm_FODA_h.cm_code);
		u.u_error = EINVAL; goto create_err;
	    }

	} else {		/* creation of a TGT	*/
	    register	i, len;

	    /* for deallocation in case of getett() returning NULL	*/
	    d->tgt.ettsz = 0;

	    /* allocate memory for timetable, copy and check it	*/
	    if ((d->tgt.cur = d->tgt.ett = getett(ctl.create.ttsz)) == NULL)
		{ u.u_error = ENOMEM; goto create_err; }
	    d->tgt.ettsz = ctl.create.ttsz;
	    for (i = 0, len = 0; i < ctl.create.ttsz-1; i++) {
		e = &(d->tgt.ett[i]);
		if (copyin(&(ctl.create.utt[i]), e, TTE_SZ))
		    { u.u_error = EFAULT; goto create_err; }
		if ((e->ete_rlen = e->ete_len) < TG_MINDPL ||
		    e->ete_len > TG_MAXDPL ||
		    (e->ete_rDt = e->ete_Dt) == 0) {
			scrivi4("%dth entry te_len = %d, te_Dt = %d\n",
				i, e->ete_len, e->ete_Dt);
			u.u_error = EINVAL; goto create_err;
		} else {
		    len = max(len, (int)(e->ete_len));
		}
	    } /* for all entries in timetable */
	    e = &(d->tgt.ett[i]);
	    if (copyin(&(ctl.create.utt[i]), e, TTE_SZ))
		{ u.u_error = EFAULT; goto create_err; }
	    if (e->ete_Dt || e->ete_len) {
		scrivi2("last entry (%dth) in timetable not null\n", i);
		u.u_error = EINVAL; goto create_err;
	    } else
		e->ete_rDt = e->ete_rlen = 0;

	    /* lock data pkt controller memory, convert pointer	*/
	    if (!userdma(ctl.create.udp, TG_MAXDPL, B_READ | B_WRITE))
		{ u.u_error = EFAULT; goto create_err; }
	    d->tgt.dp = (dp_t *)uvtophys(ctl.create.udp);

	    /* check position into click and type of data pkt	*/
	    if (btoct(d->tgt.dp) != btoct((int)d->tgt.dp + len)) {
		scrivi3("packet spans two clicks: from 0x%x to 0x%x\n",
			d->tgt.dp, (int)d->tgt.dp + len);
		u.u_error = EINVAL; goto create_err;
	    }
	    if (d->tgt.dp->dp_FODA_h.dp_code & CTL_CODE) {
		scrivi2("create: wrong code 0x%x\n",
			d->tgt.dp->dp_FODA_h.dp_code);
		u.u_error = EINVAL; goto create_err;
	    }

	    /* initialise sequence no	*/
	    d->tgt.dp->dp_MTG_h.seq_num = 0;
	}

	dp = &mtg_head;

	/* skip all TGT_D's:				*/
	/* dp points to the pointer to the first CTL_D	*/
	while ((*dp) && ((*dp)->d_type ==  TGT_D))
	    dp = &(*dp)->d_nxt;

	/* enqueue the tgt_t	*/
	d->d_nxt = (*dp);
	(*dp) = d;

	break;

	/* common exit point for errors occurred after gettgt()	*/
create_err:
	freetgt(d);
	if (cmd == TG_CREATE)
	    freeett(d->tgt.ettsz, d->tgt.ett);
	break;

    case TG_KILL:
	if (!mtg_inited)
	    { u.u_error = EINVAL; break; }
	COPYIN(arg, &ctl, sizeof(ctl.kill))
	clearcm();
	{   register dp_t	*udp = (dp_t *)uvtophys(ctl.kill.udp);
	    int			found = 0;

	    dp = &mtg_head;

	    /* check all TGT_D's and discard the right one	*/
	    while ((*dp) && ((*dp)->d_type == TGT_D)) {
		if ((*dp)->tgt.dp != udp)
		    dp = &(*dp)->d_nxt;
		else {
		    found = 1;
		    d = (*dp);
		    (*dp) = d->d_nxt;
		    while (mtg_scanindex == d)
			;
		    ctl.kill.last_seq = d->tgt.dp->dp_MTG_h.seq_num;
		    freeett(d->tgt.ettsz, d->tgt.ett);
		    freetgt(d);
		    break;
		}
	    }

	    if (!found)
		{ u.u_error = EINVAL; break; }
	}

	/* unlock data packet controller memory	*/
	undma(ctl.udp, TG_MAXDPL, B_READ | B_WRITE);

	/* give the controller the TGT last sequence number	*/
	COPYOUT(&ctl, arg, sizeof(ctl.kill))

	break;

    case TG_RATE:
	if (!mtg_inited)
	    { u.u_error = EINVAL; break; }
	COPYIN(arg, &ctl, sizeof(ctl.rate))
	clearcm();
	{   register dp_t	*udp = (dp_t *)uvtophys(ctl.rate.udp);
	    register		found = 0;

	    for (d = mtg_head; d && d->d_type == TGT_D; d = d->d_nxt)
		if (d->tgt.dp == udp) {
		    found = 1;
		    if (ctl.rate.Dt_f <= 0 || ctl.rate.len_f <= 0)
			{ u.u_error = EINVAL; break; }
		    for (e = d->tgt.ett; e->ete_Dt; e++) {
			e->ete_rDt = max((int)(ctl.rate.Dt_f * e->ete_Dt), 1);
			e->ete_rlen =
			    min(max((int)(ctl.rate.len_f * e->ete_len),
			    TG_MINDPL), TG_MAXDPL);
		    }
		    break;
		}
	    if (!found)
		{ u.u_error = EINVAL; break; }
	}
	break;

    case TG_ON:
    case TG_OFF:
	if (!mtg_inited)
	    { u.u_error = EINVAL; break; }
	COPYIN(arg, &ctl, sizeof(ctl.udp))
	clearcm();
	{   register dp_t	*udp = (dp_t *)uvtophys(ctl.udp);
	    register		found = 0;

	    for (d = mtg_head; d && d->d_type == TGT_D; d = d->d_nxt)
		if (d->tgt.dp == udp) {
		    found = 1;
		    switch (cmd) {
		    case TG_ON:
			d->d_active = 1;
			break;
		    case TG_OFF:
			d->d_active = 0;
			break;
		    }
		    break;
		}
		if (!found)
		    { u.u_error = EINVAL; break; }
	}
	break;

    default:
	u.u_error = EINVAL;
#undef	COPYIN
#undef	COPYOUT
} /* switch cmd */
} /* mtg_ioctl */
