/*
**		@(#)MTG-Ethernet:lantry.c	8.2	91/05/16
*/
#ident	"@(#)MTG-Ethernet:lantry.c	8.2	91/05/16"

#include "glob.h"
#include <sys/signal.h>		/********************************/
#include <sys/fs/s5dir.h>	/* the user structure		*/
#include <sys/user.h>		/********************************/
#include <sys/errno.h>		/* return errors for ioctl	*/


#define	lowtwobytes(a)	((ushort)(a))
#define	thirdbyte(a)	((unchar)((uint)(a) >> 16))

#define	RLEN	1	/* two's exponents of numbers of receive	*/
#define	TLEN	0	/* and transmit buffer descriptors (0-7)	*/
static struct le_md	*rring,		/* Receive Descriptor Ring	*/
			*tring;		/* Transmit Descriptor Ring	*/
static uint	currmdndx;		/* current rmd index		*/
#define		currmd	rring[currmdndx]
static rbuf_t	*currbuf;		/* current rbuf			*/

static struct tbuf {
	e_frame	t_eframe;
	unchar	t_data[ETH_MINDATA];
} tbuf;		/* transmit buffer for testing	*/
static	tbcnt;	/* number of transmitted packets	*/



void
mtg_ltryintr()
{
	register struct le_md	*Rrmd;
	register rbuf_t		*Rrbuf;
	register ushort	csr0 = LNC_rdp;

	/* check for hardware problems	*/
	if (!(csr0 & LE_INTR)) {
		(void) cmn_err(CE_WARN, "mtg_lintr: interrupt error.");
		(void) cmn_err(CE_CONT, "csr0 = 0x%x.", csr0);
error:		PCC_ticr = 0;		/* stop timer	*/
		LNC_rdp = LE_STOP;	/* stop LANCE	*/
		mtg_ok = 0;
		wakeup(&le_iblk);
		return;
	}

	/* check for software problems */
	if (LNC_rap) {
		(void) cmn_err(CE_WARN, "mtg_lintr:interrupt error.");
		(void) cmn_err(CE_CONT, "RAP = 0x%x.", LNC_rap);
		goto error;
	}

	/* check for Ethernet errors	*/
	if (csr0 & LE_ERR) {
		PCC_licr = 0;
		wakeup(&le_iblk);
		return;
	}

	/* handle Rx interrupt */
	if (csr0 & LE_RINT) do {
		Rrmd = &currmd;
		Rrbuf = currbuf;

		if (Rrmd->md_err || !Rrmd->md_stp || !Rrmd->md_enp ||
		    (Rrmd->md_mcnt != (-tring->md_bcnt + 4))) {
			PCC_licr = 0;
			wakeup(&le_iblk);
			return;
		}
		currbuf = (Rrbuf = Rrbuf->r_nxt);
		if ((Rrbuf == NULL) || (Rrbuf->r_nxt == NULL))
			break;
		Rrmd->md_hadr = thirdbyte(Rrbuf->r_nxt->r_.data);
		Rrmd->md_ladr = lowtwobytes(Rrbuf->r_nxt->r_.data);
		Rrmd->md_mcnt = 0;
		Rrmd->md_own = 1;
		currmdndx = 1 - currmdndx;	/* switch between 1 and 0 */
	} while (!currmd.md_own);

	/* handle Tx interrupt */
	if (csr0 & LE_TINT) {
		if (tring->md_err) {
			PCC_licr = 0;
			wakeup(&le_iblk);
			return;
		}
		/* here one can gather statistics on retries	*/
		/* and reset retry bits	*/
		tring->md_bcnt = -(14 + filltb(tbuf.t_data, ++tbcnt));
		tring->md_own = 1;
		csr0 |= LE_TDMD;
	}

	/* clear pending interrupt(s)	*/
	LNC_rdp = csr0;
} /* mtg_ltryintr */



static
filltb(s, nc)
register unchar	*s;
register	nc;
{
	register	i;

	/* string must be shorter than 12 chars	*/
	(void) strcpy((char *)s, "Packet n. ");

	for (i = 0; *s; s++, i++)
		;
	*s++ = "0123456789"[(nc/1000)%10];
	*s++ = "0123456789"[(nc/100)%10];
	*s++ = "0123456789"[(nc/10)%10];
	*s++ = "0123456789"[nc%10];
	*s = '\0';
	return (i + 5);
} /* filltb */



char
mtg_lantry(arg)
mtg_ctl	*arg;
{
	int	rbno;		/* number of rbuf's		*/
	rbuf_t	*rbuf;		/* array of rbno rbuf's 	*/
	register	i;

	/* stop the LANCE	*/
	mtg_lstop();

	/* check the address of initialisation block	*/
	if ((uint)&le_iblk & 0xff000001) {
	    (void) cmn_err(CE_WARN,
		    "LANCE init block allocated above the first 16Mbytes");
	    (void) cmn_err(CE_CONT, "or not word aligned.");
	    mtg_ok = 0;
	    return EFAULT;
	}

	/* Set the address of the initialisation block	*/
	LNC_rap = LE_CSR1;
	LNC_rdp = lowtwobytes(&le_iblk);
	LNC_rap = LE_CSR2;
	LNC_rdp = thirdbyte(&le_iblk);

	/* set the LANCE hardware paramemetres	*/
	LNC_rap = LE_CSR3;
	LNC_rdp = LE_BSWP | LE_BCON;

	/* reset the Register Address Port to CSR0	*/
	LNC_rap = LE_CSR0;

	/* check the address of mtg_crbuf */
	if (mtg_crbuf == NULL)
		return EINVAL;
	if ((uint)mtg_crbuf & 0xff000007) {
		(void) cmn_err(CE_WARN,
		       "Receive buffer is allocated  above the first 16Mbytes");
		(void) cmn_err(CE_CONT, "or it is not quadword aligned.");
		mtg_ok = 0;
		return EFAULT;
	}

	/* set the locations of receive and transmit descriptors rings	*/
	rring  = (struct le_md *)mtg_crbuf;
	tring  = &rring[1 << RLEN];

	/* fill the initialisation block	*/
	le_iblk.ib_prom = 0;		le_iblk.ib_intl = 0;
	le_iblk.ib_drty = 0;		le_iblk.ib_coll = 0;
	le_iblk.ib_dtcr = 0;		le_iblk.ib_loop = 1;
	le_iblk.ib_dtx  = 0;		le_iblk.ib_drx  = 0;

	swab(ETH_addr, le_iblk.ib_padr, 6);
	swab(ETH_mask, le_iblk.ib_ladrf, 8);

	le_iblk.ib_rdrp.drp_len  = RLEN;
	le_iblk.ib_rdrp.drp_drah = thirdbyte(rring);
	le_iblk.ib_rdrp.drp_dral = lowtwobytes(rring);
	le_iblk.ib_tdrp.drp_len  = TLEN;
	le_iblk.ib_tdrp.drp_drah = thirdbyte(tring);
	le_iblk.ib_tdrp.drp_dral = lowtwobytes(tring);
	
	/* initialise the LANCE	*/
	LNC_rdp = LE_INIT;

	/* zero down the receive buffer, for debugging purpose	*/
	(void) bzero(mtg_crbuf, ctob(mtg_crbsize));

	/* Set rbuf array head and count, currbuf and currmdndx	*/
	rbuf = (rbuf_t *)&tring[1 << TLEN];
	rbno = (mtg_crbuf + ctob(mtg_crbsize) - (caddr_t)rbuf) / sizeof(rbuf_t);
	if (rbno <= (1 << RLEN))
		return ENOMEM;
	currbuf = rbuf;
	currmdndx = 0;

	/* initialise the receive and transmit rings	*/
	for (i = 0; i < (1 << RLEN); i++) {
		rring[i].md_own   = 1;	rring[i].md_err   = 0;
		rring[i].md_fram  = 0;	rring[i].md_oflo  = 0;
		rring[i].md_crc   = 0;	rring[i].md_buff  = 0;
		rring[i].md_stp   = 0;	rring[i].md_enp   = 0;
		rring[i].md_mcnt  = 0;	rring[i].md_bcnt = -RBUFLEN;
		rring[i].md_hadr = thirdbyte(rbuf[i].r_.data);
		rring[i].md_ladr = lowtwobytes(rbuf[i].r_.data);
	}

	tring->md_own   = 1;	tring->md_err   = 0;
	tring->md_more  = 0;	tring->md_one   = 0;
	tring->md_def   = 0;	tring->md_stp   = 1;
	tring->md_enp   = 1;	tring->md_errflg  = 0;
	tring->md_hadr = thirdbyte(&tbuf);
	tring->md_ladr = lowtwobytes(&tbuf);

	/* set up the receive buffers chain	*/
	{	register rbuf_t	*b;
	
		for (b = rbuf; b < rbuf + rbno - 1; b++)
			b->r_nxt = b + 1;
		b->r_nxt = NULL;
	}

	/* fill the transmit buffer	*/
	(void) bcopy((caddr_t)ETH_addr, (caddr_t)tbuf.t_eframe.src, 6);
	(void) bcopy((caddr_t)ETH_addr, (caddr_t)tbuf.t_eframe.dest, 6);
	tbuf.t_eframe.type = ETYP_IP;
	tring->md_bcnt = -(14 + filltb(tbuf.t_data, (tbcnt = 0)));

	/* wait for initialisation done with timeout	*/
	delay(2);
	if (LNC_rdp != (LE_IDON | LE_INTR | LE_INIT)) {
		(void) cmn_err(CE_WARN, "LANCE error during initialisation.");
		(void) cmn_err(CE_CONT, "csr0 = 0x%x.");
		mtg_ok = 0;
		return EIO;
	}

	/* clear the interrupt condition, set the interrupt vector	*/
	/* and start the LANCE	*/
	LNC_rdp = LE_IDON;
	M68Kvec[LANVECT] = (int(*)())mtg_ltryint;
	PCC_licr = PCC_IEN | PCC_IL1;
	LNC_rdp = LE_INEA | LE_STRT;
 	
	/* wait for rbuf's exausted, then return csr0	*/
	if (sleep(&le_iblk, PCATCH | PUSER))
		return EINTR;
	if (sushort((ushort *)arg, LNC_rdp) < 0)
		return EFAULT;

	/* return successfully	*/
	return 0;
} /* mtg_lantry */
