/*
**		@(#)MTG-Ethernet:low.c	9.7	92/11/19
*/
#ident	"@(#)MTG-Ethernet:low.c	9.7	92/11/19"

#include "glob.h"


#define	heartbeat_test		1	/* 0 to ignore CERR in csr0	*/
#define	LANCE_consistency_check	0	/* 0 to bypass check		*/
#define	Rx_unespected_int	0	/* 0 to bypass check		*/
#define	Tx_unespected_int	0	/* 0 to bypass check		*/

#define	lowtwobytes(a)	((ushort)(a))
#define	thirdbyte(a)	((unchar)((uint)(a) >> 16))

static int		phystokv_off;	/* offset between physical and	*/
					/* kernel virtual addresses	*/
static struct le_md	*rring,		/* physical and virtual addr. of*/
			*vrring,	/* Receive Descriptor Ring	*/
			*tring,		/* physical and virtual addr. of*/
			*vtring;	/* Transmit Descriptor Ring	*/
static uint	currmdndx;		/* current rmd index		*/
#define		currmd	vrring[currmdndx]
#define		curtmd	vtring[0]
static ebuf_t	*curebuf;		/* current error message	*/
static rbuf_t	*currbuf,		/* phys. and virt.(not cached)	*/
		*curvrbuf;		/* addr. of curr. receive buffer*/

static struct cmpp {		/*** control message pointer pointer ****/
	struct cmpp	*cmpp_nxt;	/* forward pointer		*/
	caddr_t		*cmpp_p;	/* phys.addr. of controller's cmpp_p */
}	cmpa[CMP_SIZE],			/* array of ptrs to controller's cmp */
	*curcmpp;			/* current cmp pointer		     */

static struct dppp {		/***** data packet pointer pointer ******/
	struct dppp	*dppp_nxt;	/* forward pointer		*/
	caddr_t		*dppp_p;	/* phys.addr. of controller's dpp_p  */
	uint		*dppp_sz;	/* phys.addr. of controller's dpp_sz */
}	dppa[DPP_SIZE],			/* array of ptrs to controller's dpp */
	*curdppp;			/* current dpp pointer		     */


#define	 FATAL(retval, errmsg, val)					\
{									\
	mtg_tstop();							\
	mtg_lstop();							\
	mtg_scanindex = INVALID_TGT;					\
	mtg_ok = 0;							\
	(void) cmn_err(CE_WARN, (errmsg), (val));			\
	return retval;							\
}



static void
error(err)
unchar	err;
{
	register ebuf_t	*Rebuf = curebuf;

	/* store error type and timestamp	*/
	Rebuf->err_msg.cm_FODA_h.cm_code = err;
	Rebuf->err_msg.cm_MTG_h.mtx_time = *mtg_utick;

	/* fill current controller cmp slot and rbuf back pointer */
	*(Rebuf->err_back = curcmpp->cmpp_p) = Rebuf->err_uaddr;
	
	/* advance current cmpp and ebuf pointers	*/
	curcmpp = curcmpp->cmpp_nxt;
	curebuf = Rebuf = Rebuf->err_nxt;

	/* the last controller cmp_p that has been pointing to the next	*/
	/* ebuf must now be NULLed or even point  somewhere else.	*/
	if ((Rebuf->err_back) && (*Rebuf->err_back == Rebuf->err_uaddr))
		FATAL(, TF_ERRE, NULL)

	/* check next controller cmp slot	*/
	if (*curcmpp->cmpp_p != NULL)
		FATAL(, TF_CMPE, NULL)
} /* error */



/*
 * Tick timer initialiser (called from mtg_ioctl())
 */
char
mtg_tinit(tick)
time_t	*tick;
{
	register time_t	tmp;
	register	i;

	/* the tick will be called at utick multiples of 10	*/
	*mtg_utick = -1;
	wait_for_tick = 1;
	for (i = 0; i < 16; utick_del[i++] = 0)
		;
	if ((*tick < 500) || (*tick > 5*1000*1000))
		return EINVAL;
	tmp = *tick * 100;		/* forcing order of evaluation	*/
	tmp /= TICK_RATIO;
	PCC_tpr = 0x10000l - tmp / 625; /* preload value		*/
	tmp = (0x10000l - PCC_tpr) * 625;
	tmp *= TICK_RATIO;
	*tick = tmp / 100;		/* (truncated) real tick length	*/

	PCC_tcr = PCC_CLROVF;		/* stop, preload, clear overflow */
	PCC_tcr = PCC_ENAB | PCC_CTRENA;/* start counting		*/
	PCC_ticr = PCC_IEN | PCC_IL7;	/* enable level 7 interrupts	*/

	return 0;
} /* mtg_tinit */



/*
 * LANCE initialiser (called from mtg_ioctl())
 */
char
mtg_linit(ucrbuf)
caddr_t	ucrbuf;
{
ebuf_t	*ebuf;			/* array of error messages	*/
rbuf_t	*rbuf;			/* array of receive buffers	*/
register i;

    /* stop the LANCE	*/
    mtg_lstop();

    /* check the address of the initialization block	*/
    if ((uint)&le_iblk & 0xff000001)
	FATAL(EFAULT, TF_LNCA, NULL)

    /* calculate the offset between physical and kernel virtual addresses*/
    phystokv_off = (int)mtg_vcrbuf - (int)mtg_crbuf;

    /* set the address of the initialisation block	*/
    LNC_rap = LE_CSR1;
    LNC_rdp = lowtwobytes(&le_iblk);
    LNC_rap = LE_CSR2;
    LNC_rdp = thirdbyte(&le_iblk);

    /* set the LANCE hardware parametres	*/
    LNC_rap = LE_CSR3;
    LNC_rdp = LE_BSWP | LE_BCON;

    /* reset the Register Address Port to CSR0	*/
    LNC_rap = LE_CSR0;

    /* check the address of mtg_lanrbuf	*/
    if ((uint)mtg_lanrbuf & 0xff000007)
	FATAL(EFAULT, TF_RCVB, NULL)

    /* set the locations of receive and transmit descriptors rings,	*/
    /* and their virtual cache inhibited  addresses			*/
    rring  = (struct le_md *)mtg_lanrbuf;
    tring  = &rring[1 << RLEN];
    vrring = (struct le_md *)((int)rring + phystokv_off);
    vtring = (struct le_md *)((int)tring + phystokv_off);

    /* fill the initialisation block	*/
    le_iblk.ib_prom = 1;		le_iblk.ib_intl = 0;
    le_iblk.ib_drty = 0;		le_iblk.ib_coll = 0;
    le_iblk.ib_dtcr = 0;		le_iblk.ib_loop = 0;
    le_iblk.ib_dtx  = 0;		le_iblk.ib_drx  = 0;

    swab(ETH_addr, le_iblk.ib_padr, 6);
    swab(ETH_mask, le_iblk.ib_ladrf, 8);

    le_iblk.ib_rdrp.drp_len  = RLEN;
    le_iblk.ib_rdrp.drp_drah = thirdbyte(rring);
    le_iblk.ib_rdrp.drp_dral = lowtwobytes(rring);
    le_iblk.ib_tdrp.drp_len  = TLEN;
    le_iblk.ib_tdrp.drp_drah = thirdbyte(tring);
    le_iblk.ib_tdrp.drp_dral = lowtwobytes(tring);

    /* initialise the LANCE	*/
    LNC_rdp = LE_INIT;

    {	register uint		rbno;
	register uint		vtopdiff = ucrbuf - mtg_crbuf;
	register caddr_t	cp;

	/* Set error messages array head	*/
	cp = (caddr_t)&tring[1 << TLEN];
	(void) bcopy("<-------- Error buffers ------->", cp, 32);
	ebuf = (ebuf_t *)(cp + 32);

	/* set up the error message chain and initialise curebuf	*/
	for (i = 0; i < ERR_SIZE; i++) {
	    ebuf[i].err_nxt   = &ebuf[i+1];
	    ebuf[i].err_uaddr = (caddr_t)&ebuf[i].err_msg +vtopdiff;
	    ebuf[i].err_back  = NULL;
	}
	curebuf = ebuf[i-1].err_nxt = ebuf;

	/* Set receive buffers array head and count	*/
	cp = (caddr_t)&ebuf[ERR_SIZE];
	(void) bcopy("<------- Receive buffers ------>", cp, 32);
	rbuf = (rbuf_t *)(cp + 32);
	rbno = ((int)mtg_crbuf + ctob(mtg_crbsize) - (int)rbuf) /sizeof(rbuf_t);
	if (rbno <= (1 << RLEN))
	    return ENOMEM;
	if (rbno > max(CMP_SIZE, DPP_SIZE))
	    (void) cmn_err(CE_NOTE,
			     "number of buffers > max(CMP_SIZE, DPP_SIZE)");

	/* set physical and kernel virtual currbuf and currmdndx	*/
	currbuf = rbuf;
	curvrbuf = (rbuf_t *)((int)rbuf + phystokv_off);
	currmdndx = 0;

	/* set up the receive buffers chain	*/
	for (i = 0; i < rbno; i++) {
	    rbuf[i].r_nxt     = &rbuf[i+1];
	    rbuf[i].r_rlennxt = &rbuf[(i+(1<<RLEN)) % rbno];
	    rbuf[i].r_uaddr   = rbuf[i].r_.data + vtopdiff;
	    rbuf[i].r_back    = NULL;
	}
	rbuf[i-1].r_nxt = rbuf;
    }

    /* initialise the receive and transmit rings	*/
    for (i = 0; i < (1 << RLEN); i++) {
	rring[i].md_own   = 1;	rring[i].md_err   = 0;
	rring[i].md_fram  = 0;	rring[i].md_oflo  = 0;
	rring[i].md_crc   = 0;	rring[i].md_buff  = 0;
	rring[i].md_stp   = 0;	rring[i].md_enp   = 0;
	rring[i].md_mcnt  = 0;	rring[i].md_bcnt = -RBUFLEN;
	rring[i].md_hadr = thirdbyte(rbuf[i].r_.data);
	rring[i].md_ladr = lowtwobytes(rbuf[i].r_.data);
    }

    tring->md_own  = 0;	tring->md_err    = 0;
    tring->md_more = 0;	tring->md_one    = 0;
    tring->md_def  = 0;	tring->md_stp    = 1;
    tring->md_enp  = 1;	tring->md_errflg = 0;

    /* reset the retry counters	*/
    mtg_stat.st_tot = mtg_stat.st_def = 0;
    mtg_stat.st_one = mtg_stat.st_more = mtg_stat.st_rtry = 0;

    /* initialise cmpa & dppa, curcmpp & curdppp	*/
    for (i = 0; i < CMP_SIZE; i++) {
	cmpa[i].cmpp_nxt = &cmpa[i+1];
	cmpa[i].cmpp_p = (caddr_t *)uvtophys(&mtg_ucmp_a[i].cmp_p);
    }
    curcmpp = cmpa[i-1].cmpp_nxt = cmpa;

    for (i = 0; i < DPP_SIZE; i++) {
	dppa[i].dppp_nxt = &dppa[i+1];
	dppa[i].dppp_p = (caddr_t *)uvtophys(&mtg_udpp_a[i].dpp_p);
	dppa[i].dppp_sz = (uint *)uvtophys(&mtg_udpp_a[i].dpp_sz);
    }
    curdppp = dppa[i-1].dppp_nxt = dppa;

    /* wait for initialization done with timeout	*/
    delay(2);
    if (LNC_rdp != (LE_IDON | LE_INTR | LE_INIT))
	FATAL(EIO, TF_LNCI, LNC_rdp)

    /* clear the interrupt condition and start the LANCE	*/
    LNC_rdp = LE_IDON;
    PCC_licr = PCC_IEN | PCC_IL7;
    LNC_rdp = LE_INEA | LE_STRT;

    /* return successfully	*/
    return 0;
} /* mtg_linit */



/*
 * Stops Tick timer activity
 */
void
mtg_tstop()
{
    PCC_ticr = 0;	/* disable interrupts	*/
} /* mtg_tstop */



/*
 * Stops LANCE activity
 */
void
mtg_lstop()
{
    PCC_licr = 0;	/* disable interrupts	*/
    LNC_rap = LE_CSR0;	/*    stop	*/
    LNC_rdp = LE_STOP;	/*  the LANCE	*/
} /* mtg_lstop */



/*
 *   Scans the TGT linked list and decrements the semaphore at the end of the
 * list, then rescans from the head of the list if the semaphore is not
 * negative. mtg_scan() is called from mtg_tint() (assembler) and mtg_lintr().
 * mtg_scan() uses always physical addresses when it accedes the rings and
 * the receive buffers, since it always writes and never reads, so there is
 * no cache consistency problem.
 */
void
mtg_scan(tdp)
register tgt_t	*tdp;
{
    register cm_t	*cmdata;
    register dp_t	*dpdata;
    register ett_e	*curett;

#if mc68000
	asm ("	scan_semaphore_loop:		");
#else
    do {	/* while semaphore >= 0	*/
#endif
	for (; tdp != NULL; tdp = tdp->d_nxt) {
	    if (tdp->d_active && (--tdp->d_counter == 0)) {
		switch (tdp->d_type) {
		case CTL_D:			/* ctl message descriptor */
		    cmdata = &tdp->ctl.ctl;

		    /* set address and length of packet to be sent	*/
		    tring->md_hadr = thirdbyte(cmdata);
		    tring->md_ladr = lowtwobytes(cmdata);
#ifndef lint
		    tring->md_bcnt = ((TG_CTLL > ETH_MIN_TU) ?
			    -(short)TG_CTLL : -(short)ETH_MIN_TU);
#endif

		    /* inactivate ctl message descriptor	*/
		    tdp->d_active = 0;

		    /* set MTG Tx timestamp	*/
		    cmdata->cm_MTG_h.mtx_time = *mtg_utick;
		    break;

		case TGT_D:			/* TGT 		*/
		    dpdata = tdp->tgt.dp;
		    curett = tdp->tgt.cur;

		    /* set address and length of packet to be sent	*/
		    tring->md_hadr = thirdbyte(dpdata);
		    tring->md_ladr = lowtwobytes(dpdata);
		    tring->md_bcnt = -curett->ete_rlen;

		    /* set MTG sequence number and FODA length	*/
		    dpdata->dp_MTG_h.seq_num += 1;
		    dpdata->dp_FODA_h.dp_len =
			    curett->ete_rlen - sizeof(e_frame) - FODA_DATAHL;

		    /* advance current ett pointer	*/
		    tdp->d_counter = curett->ete_rDt;
		    tdp->tgt.cur=((++curett)->ete_rDt) ? curett : tdp->tgt.ett;

		    /* set MTG Tx timestamp	*/
		    dpdata->dp_MTG_h.mtx_time = *mtg_utick;
		    break;
		default:
		    FATAL(, TF_SCCF, NULL)
		}
		/* update the ''true'' pointer	*/
		mtg_scanindex = tdp;

		/* make the LANCE run	*/
		tring->md_own = 1;
		LNC_rdp = LE_INEA | LE_TDMD;
		/* only now I can interrupt myself	*/

		return;
	    } /* if (DMA must be initiated) */
	} /* for (all tgt_t in linked list)	*/
	mtg_scanindex = INVALID_TGT;
	tdp = mtg_head;

#if mc68000
	asm ("	sub.l	&1,([mtg_sem])		");
	asm ("	bge	scan_semaphore_loop	");
#else
    } while (--(*mtg_sem) >= 0);
#endif
} /* mtg_scan */



/*
 * LANCE handler
 */
void
mtg_lintr()
{
    register ushort	csr0 = LNC_rdp;
    register		eth_error;
    register rbuf_t		*Rrbuf;		/************************/
    register struct le_md	*Rrmd;		/*   all used for Rx	*/
    register struct cmpp	*Rcmpp;		/*	interrupt	*/
    register struct dppp	*Rdppp;		/************************/


    /* check for LANCE errors	*/
#   if (heartbeat_test)
#	define	ERR_MASK	(LE_BABL | LE_CERR | LE_MISS | LE_MERR)
#   else
#	define	ERR_MASK	(LE_BABL | LE_MISS | LE_MERR)
#   endif
    if (csr0 & (LE_IDON | ERR_MASK))
	FATAL(, TF_LNCL, csr0)

    /* Rx interrupt ?	*/
    if (csr0 & LE_RINT) {

#if (Rx_unespected_int)
	/* check for unespected interrupt	*/
	if (currmd.md_own)
	    FATAL(, TF_UNRX, currmdndx)
#endif

	do {
	    Rrbuf = currbuf;
	    Rrmd = &currmd;

	    /* check for Ethernet errors	*/
	    if (eth_error = (Rrmd->md_err)) {
		if (Rrmd->md_oflo)
		    FATAL(, TF_OFLO, NULL)
		if (Rrmd->md_buff)
		    FATAL(, TF_BUFF, NULL)
	        Rrmd->md_err = 0;
		Rrmd->md_fram = 0;
		Rrmd->md_crc = 0;
	    }

	    if (curvrbuf->r_.cm_data.cm_FODA_h.cm_code & CTL_CODE) {
	    /*********************** ctl message ***********************/
		Rcmpp = curcmpp;

		/* set timestamp	*/
		Rrbuf->r_.cm_data.cm_MTG_h.mrx_time = *mtg_utick;

		/* set receive error	*/
		Rrbuf->r_.cm_data.cm_MTG_h.frx_err = eth_error;

		/* fill current controller cmp slot and rbuf back pointer */
		*(Rrbuf->r_back = Rcmpp->cmpp_p) = Rrbuf->r_uaddr;
			    
		/* advance current cmpp pointer	*/
		curcmpp = Rcmpp = Rcmpp->cmpp_nxt;

		/* check next controller cmp slot	*/
		if (*Rcmpp->cmpp_p != NULL)
		    FATAL(, TF_CMPL, NULL)
	    } else {
	    /*********************** data packet ***********************/
		Rdppp = curdppp;

		/* set timestamp	*/
		Rrbuf->r_.dp_data.dp_MTG_h.mrx_time = *mtg_utick;

		/* set receive error	*/
		Rrbuf->r_.dp_data.dp_MTG_h.frx_err = eth_error;

		/* fill current controller cmp slot and rbuf back pointer */
		*(Rrbuf->r_back = Rdppp->dppp_p) = Rrbuf->r_uaddr;
		*Rdppp->dppp_sz = Rrmd->md_mcnt - 4;
		    
		/* advance current dppp pointer	*/
		curdppp = Rdppp = Rdppp->dppp_nxt;

		/* check next controller dpp slot	*/
		if (*Rdppp->dppp_p != NULL)
		    FATAL(, TF_DPPL, NULL)
	    }

	    /* advance physical and kernel virtual current pointer	*/
	    curvrbuf = (rbuf_t *)((int)(currbuf = Rrbuf->r_nxt) + phystokv_off);
	    
	    /* advance next rbuf pointer	*/
	    Rrbuf = Rrbuf->r_rlennxt;

	    /* the last controller cmp_p or dpp_p that has been */
	    /* pointing to rbuf after the next one must now be	*/
	    /* NULLed or even point  somewhere else.		*/
	    if ((Rrbuf->r_back) && (*Rrbuf->r_back == Rrbuf->r_uaddr)) {
		/* stop all transmit and receive activity	*/
		mtg_tstop();
		mtg_lstop();
		mtg_scanindex = INVALID_TGT;
		mtg_inited = 0;

		/* send error to controller (receive buffer overflow)	*/
		error(TE_SLOW);
		return;
	    }

	    /*   fill the current rmd with the address of the rbuf	*/
	    /* that comes after the next one, and advance current	*/
	    /* rmd pointer						*/
	    Rrmd->md_hadr = thirdbyte(Rrbuf->r_.data);
	    Rrmd->md_ladr = lowtwobytes(Rrbuf->r_.data);
	    Rrmd->md_mcnt = 0;
	    Rrmd->md_own = 1;
	    currmdndx = (currmdndx + 1) & ((1 << RLEN) - 1);
	} while (!currmd.md_own);
    } /* Rx interrupt */

    /* Tx interrupt ?	*/
    if (csr0 & LE_TINT) {

#if (LANCE_consistency_check)
	/* LANCE Tx interrupt  consistency check	*/
	if (*mtg_sem < 0)
	    FATAL(, TF_LICF, *mtg_sem)
#endif

#if (Tx_unespected_int)
	/* check for unespected interrupt	*/
	if (curtmd.md_own)
	    FATAL(, TF_UNTX, NULL)
#endif

	/* semaphore too big ?	*/
	if ((mtg_maxsem) && (*mtg_sem > mtg_maxsem)) {
	    mtg_maxsem = 0;
	    error(TE_SEMA);
	    if (!mtg_ok)	/* did error() generate a fatal error?	*/
		return;
	}

	/* gather statistics	*/
	++mtg_stat.st_tot;
	if (((char *)(&curtmd))[2] & 0x1c)
	    if (curtmd.md_def)
		{ curtmd.md_def = 0; ++mtg_stat.st_def; }
	    else if (curtmd.md_one)
		{ curtmd.md_one = 0; ++mtg_stat.st_one; }
	    else if (curtmd.md_more)
		{ curtmd.md_more = 0; ++mtg_stat.st_more; }

	/* check for errors	*/
	if (curtmd.md_err) {
	    if (curtmd.md_errflg & TMD_UFLO)
		FATAL(, TF_UFLO, NULL)
	    if (curtmd.md_errflg & TMD_LCOL)
		error(TE_LCOL);	/* late collision */
	    if (curtmd.md_errflg & TMD_LCAR)
		error(TE_LCAR);	/* loss of carrier */
	    if (curtmd.md_errflg & TMD_RTRY) {
		++mtg_stat.st_rtry;
		error(TE_RTRY);	/* retry error */
	    }
	    /* did error() generate a fatal error?	*/
	    if (!mtg_ok)
		return;

	    curtmd.md_err = 0;
	    curtmd.md_errflg = 0;
	}
	if (curtmd.md_errflg & TMD_BUFF)
	    FATAL(, TF_BUFE, NULL)

	/************** clear interrupt condition ***************/
	/* The level 7 interrupt is non-maskable. Anyway,	*/
	/* since this means that the CPU is edge-sensitive on	*/
	/* level 7 interrupts, another level 7 int can't be	*/
	/* recognised if we don't clear the current interrupt	*/
	/* condition. So we have been truly uninterruptible	*/
	/* till now.						*/
	/* The interrupt condition is cleared only now if there	*/
	/* is a pending Tx interrupt. This in order to avoid	*/
	/* race conditions on curcmpp if there were a transmit	*/
	/* error to be notified to the controller - so let's	*/
	/*  wait until error checking is done.			*/
	/********************************************************/
	LNC_rdp = csr0;

	/* scan the TGT linked list	*/
	mtg_scan(mtg_scanindex->d_nxt);

    } else { /* not Tx interrupt */

	/* clear interrupt condition	*/
	LNC_rdp = csr0;

    } /* not Tx interrupt */
} /* mtg_lintr */
