/*
**	Structures for ioctl to mtg driver.
**		%Z%%Y%%M%	%I%	%E%
*/

#ident	"%Z%%Y%%M%	%I%	%E%"

#include <sys/types.h>
#include <local/lance.h>

/************************************************************************/
/* WARNING: if the dummy driver is running, you must set mtg_dummy to 1	*/
#define	mtg_dummy	0	/****************************************/
/************************************************************************/

/*
 *   The driver can be opened by one process only at a time. EACCES error will
 * returned by open(2) whan a process will try to access the driver and 
 * another process is already accessing it. EPERM error will be returned by
 * open(2) if the process has not super-user permissions.
 */

/* Names and minor numbers for the driver special files	*/
#define	DEFAULT_DEV	"/dev/mtg"
#define	TEST_DEV	"/dev/mtgtest"
#define	DEFAULT_MIN	0
#define	TEST_MIN	1

/************************************************************************/
/*				FODA constants				*/
/************************************************************************/
/*                                                                      */
/*	 
 *	control messages from the GATEWAY toward the FODA system
 *		GATEWAY ============> TDMA CONTROLLER
 */
#define	STR_CH_REQ	0x83	/* strem channels request		*/
#define	STR_CH_REP	0x87	/* stream channels modification repl	*/
#define	STR_CH_REL	0x89	/* stream channels relinquish		*/
#define	BUILD_GR	0xA0	/* build group of users			*/
#define	CANCEL_GR	0xB0	/* cancel group of users		*/
#define	MODIFY_GR	0x8A	/* modify group of users		*/

/*
 *	control messages from the FODA system toward the GATEWAY 
 *		FODA system ========> GATEWAY
 */
#define	STR_CH_GRANT	0x81	/* stream channels granted/refused	*/
#define	STR_CH_MODIF	0x85	/* stream channels modification request	*/
#define	STOP_DATAGRAM	0x98	/* stop the sending of datagram packets	*/
#define	RESUME_DATAG	0x88	/* resume sending of datagram packets	*/
#define	DST_NOT_RCH	0xD0	/* destination not reachable		*/
#define	DST_NOW_RCH	0xC1	/* destination now reachable		*/
#define	BLD_GR_REP	0xC2	/* build a group of users reply		*/
#define	MOD_GR_REP	0xC3	/* extend a users group reply		*/
#define	REQ_ERROR	0xC4	/* error in request			*/

/* All the control messages have this bit on */
#define	CTL_CODE	0x80	/* control messages code bit */

/* All the data messages can be or'ed with these */
#define MTG_CODE	0x40	/* MTG-generated code bit */
#define CRC_CODE	0x20	/* CRC-requested code bit */

/*	 
 *	data messages from the GATEWAY toward the FODA system
 *		GATEWAY ============> FODA SYSTEM
 */
#define	SND_BULK	0x11	/* send bulk data without CRC		*/
#define	SND_INT		0x12	/* send interactive data without CRC	*/
#define	SND_STR		0x13	/* send stream data without CRC		*/

/*
 *	data messages from the FODA system toward the GATEWAY 
 *		FODA system ========> GATEWAY
 */
#define	RCV_BULK	0x01	/*  bulk data without CRC received      */
#define	RCV_INT		0x02	/*  interact. data without CRC received	*/
#define	RCV_STR		0x03	/*  stream data without CRC received    */

/* STR_CH_REQ broadcast constant */
#define GAFO_BROADCAST	255

/* Ethernet type for GAFO packets */
#define GAFO_TYPE	0

/*
 * time frame
 */
#define	TIME_FRAME	20.0		/* length of a frame (msec)	*/
#define	EL_SLOT		4		/* size of a slot (bytes)	*/
#define	EL_SLOT_B	(EL_SLOT * 8)	/* size of a slot (bits)	*/

/****************************************************************/
/*	Data structures common to controller and driver		*/
/****************************************************************/

/* Size of data structures common to controller and driver	*/
#define	CMP_SIZE	50		/* size of CMP array		*/
#define	DPP_SIZE	4200		/* size of DPP array		*/
					/* must be about 700 per MByte	*/

/* Length of various data structures	*/
/* minimum and maximum packet lengths	*/
#	define	TG_MINDPL	ETH_MIN_TU
#if (mtg_dummy)
#	define	TG_MAXDPL	150
#else
#	define	TG_MAXDPL	ETH_MAX_TU
#endif
/* length of FODA control message and data packet headers	*/
#define	FODA_CTLHL	sizeof(struct fctl_hdr)
#define	FODA_DATAHL	sizeof(struct fdata_hdr)
/* length of FODA header	*/
#define	MTG_HDRL	sizeof(tghdr_t)
/* maximum size of data in a data packet	*/
#define	TG_MAXDATA	(TG_MAXDPL - sizeof(e_frame) - FODA_DATAHL - MTG_HDRL)
/* length of a FODA control message	*/
#define	TG_CTLL		(sizeof(cm_t))


/*
 * MTG  header format
 */
typedef struct tg_hdr {
	 ushort seq_num;	/* seq.number set by MTG-Eth.		*/
	 ushort	tgID	:8;	/* Traffic Generator ID			*/
	 ushort	monitor	:1;	/* monitor bit set by MTG_Eth.		*/
	 ushort	ftx_err	:1;	/* XMT error from Ethernet		*/
	 ushort	frx_err	:1;	/* RCV error from Ethernet		*/
	 ushort		:5;	/* reserved				*/
	 unchar	EbN0;		/* Signal/Noise ratio set by satellite	*/
	 unchar	bc_rate;	/* bit and coding rate set by FODA	*/
	 unchar	ftx_lev;	/* XMT signal level set by FODA		*/
	 unchar	frx_lev;	/* RCV signal level set by FODA		*/
	 ushort	queuel;		/* backlog				*/
	 time_t mtx_time;	/* XMT timestamp set by MTG-Eth.	*/
	 time_t ftx_time;	/* RCV timestamp set by FODA-Eth. UPside*/
	 time_t stx_time;	/* XMT timestamp set by FODA-Eth. UPside*/
	 time_t srx_time;	/* RCV timestamp set by FODA-Eth. DNside*/
	 time_t frx_time;	/* XMT timestamp set by FODA-Eth. DNside*/
	 time_t mrx_time;	/* RCV timestamp set by MTG-Eth.	*/
         ushort quality;        /* quality count (datagram only)	*/
} tghdr_t;

/*
 * Traffic Generator timetable entry
 */
typedef struct {
	ushort	te_Dt;		/* ticks from previous entry to this one */
	ushort	te_len;		/* length of this packet in bytes	*/
} tt_e;
#define	TTE_SZ	sizeof(tt_e)

/*
 * MTG ctl and error message structure
 */
typedef struct ctlm {
	e_frame		cm_Eth_h;	/* Ethernet header	*/
	struct fctl_hdr {		/* GAFO header for control messages */
		unchar cm_code;   	/* message type 		*/
		unchar cm_addr;		/* destination address		*/
		unchar cm_ugid;		/* user group identifier	*/
		unchar cm_cl;		/* requested class of service	*/
		/*unchar cm_nbw; 	 # of words of data		*/
		unchar cm_gid;		/* id assigned by gateway	*/
		unchar cm_fid; 		/* id assigned by FODA		*/
		ushort cm_chrq;		/* requested stream channels	*/
		ushort cm_chac;		/* min acceptable # of stream ch. */
		}	cm_FODA_h;		/* GAFO header		*/
	tghdr_t		cm_MTG_h;		/* MTG header		*/
} cm_t;

/*
 * MTG data packet structure
 */
typedef struct datap{
	e_frame		dp_Eth_h;	/* Ethernet header	        */
	struct fdata_hdr {		/* GAFO header for data packets	*/
		unchar dp_code; 	/* message type			*/
		unchar dp_addr;		/* destination address		*/
		unchar dp_uuid;		/* user group identifier	*/
#define                dp_gid dp_uuid   /* identifier assigned by gateway */
		unchar dp_cl;		/* class of service		*/
#define		       dp_fid dp_cl	/* identifier assigned by FODA	*/
		ushort dp_len;		/* length of the following data	*/
		}	dp_FODA_h;		/* GAFO header		*/
	tghdr_t		dp_MTG_h;		/* MTG header		*/
	unchar		dp_data[TG_MAXDATA];
#ifdef	_MTG_USER
	tt_e		*ptr_table;	/* pointer to timetable */
	tgthdr		dp_tgt;		/* TG parameters	*/
#endif
} dp_t;

/* status byte */
#define dp_sts dp_cl
#define STS_CRC		1		/* wrong satellite CRC */
#define STS_INC		2		/* incomplete packet */

/*
 *   The CMP and DPP arrays are allocated by the controller. Their address is
 * made known to the driver by use of TG_INIT. They are organized as two
 * circular queues. Size is fixed at compile time.
 */

/* Control Message Pointer entry	*/
typedef struct CMP_e {
	struct CMP_e	*cmp_nxt;	/* forward pointer		*/
	cm_t		*cmp_p;		/* pointer to Control Message	*/
} cmp_e;

/* Data Packet Pointer entry		*/
typedef struct DPP_e {
	struct DPP_e	*dpp_nxt;	/* forward pointer		*/
	dp_t		*dpp_p;		/* pointer to Data Packet	*/
	uint		dpp_sz;		/* length of Data Packet	*/
} dpp_e;

/****************************************************************/
/*		ioctl(2) interface with the driver.		*/
/****************************************************************/

/*
 *   These are the commands available with ioctl(2). The third argument points
 * to the union mtg_ctl. All the commands return EINVAL if the protocol is not
 * respected.
 *************** contiguous receive buffer management commands ****************
 * TG_GETRBUF
 *	A request is made to the driver to allocate a physically contiguous
 *	area as receive buffer. Only one such buffer at a time can exist in the
 *	system, so the call will fail (EINVAL) if a buffer had been already
 *	allocated and has not been released with TG_FREERBUF (see below). The
 *	buffer will be cache inhibited. The
 *	requested buffer size in clicks must be put in mtg_ctl.crb.size, while
 *	mtg_ctl.crb.physaddr must be put equal to NULL. The mtg_ctl.crb
 *	structure will contain the address of the allocated buffer
 *	and its size in clicks, or zeroes if no buffer has been allocated.
 *	<sys/types.h>, <sys/param.h> and <sys/sysmacros.h> should be included
 *	by the caller, who will use the macros ctob(clicks) and btoc(bytes)
 *	to convert between clicks and bytes. The call will fail (ENOMEM) if not
 *	enough contiguous memory is available, due to memory fragmentation or
 *	size too big. The kernel variable availrmem contains the free number
 *	of clicks available in the system. After login, the memory is usually
 *	fragmented, so a big buffer must be allocated at startup time.
 * TG_ASKRBUF
 *	A request is made to the driver to give the address and size
 *	of the allocated buffer. The parameters are returned in the mtg_ctl.crb
 *	structure. They will both be zero if no buffer is allocated.
 * TG_FREERBUF
 *	A request is made to the driver to release the physically contiguous
 *	buffer previously allocated with TG_GETREBUF (see above). The call
 *	will fail (EINVAL) if no buffer is currently allocated.
 ***************** test commands (authorized personnel only) ******************
 * TG_RESET
 *	Puts the driver in the state it has after machine boot. This command
 *	is accepted only with a minor device number equal to TEST_MIN.
 ****************** normal controller interaction commands ********************
 * TG_INIT
 *	This command initializes all the driver MTG data structures.
 *	The controller must have allocated the contiguous receive buffer by
 *	means of TG_GETRBUF, and must have mapped it in its own address
 *	space.  The addresses of the crbuf (controller address), and of the
 *	CMP and DPP arrays are passed to the driver in the mtg_ctl.init
 *	structure.  The CMP and DPP arrays must be the correct size (see
 *	below) and all their members must be longword aligned.  The first
 *	structure members are provided by the controller, while the last
 *	ones are written by the driver and read by the controller upon
 *	return from the ioctl(). tick_len is the requested tick length in
 *	usec upon call of ioctl(). It will contain the real tick length
 *	upon return from ioctl(). An EINVAL error will be returned if
 *	tick_len drops out of the interval [500 - 5000000]. An ENOSPC error
 *	is returned if the specified timetable pool size exceeds the size
 *	of the contiguous buffer, while EINVAL is returned if it is less
 *	than 1000.
 * TG_HALT
 *	This command releases the lock on the CMP and DPP arrays. TG_INIT
 *	must be issued again before calling any of the other normal controller
 *	interaction commands.
 *
 *	All the subsequent commands return error (EINVAL) if called before
 *	TG_INIT, or after TG_HALT.
 * TG_CTL
 *	The ctl message pointed by the ucm pointer in the mtg_ctl union
 *	is enqueued and sent as soon as possible. The controller can dispose
 *	the space ocuupied by the control message after this command. A
 *	ENOSPC error is returned to the controller if the driver constant
 *	TGTPOOL_SZ is too small.
 * TG_CREATE
 *	A TGT is created with parameters: activity=ON, next seq.no.=1.  The
 *	controller must supply the address of the TG data packet and the
 *	address of the TG timetable, together with its length (# of
 *	entries including the last entry).  The last entry of the timetable
 *	must have null members.  The used length of the TG data packet
 *	(the used length may be less than TG_MAXDPL) MUST NOT CROSS a click
 *	BOUNDARY.  The data packet MUST BE KEPT IN PLACE UNTIL A TG_KILL
 *	COMMAND IS ISSUED.  The controller can deallocate the space used
 *	for the timetable after this command.
 *	A ENOSPC error will occur if the driver constant TGTPOOL_SZ is too
 *	small. A ENOMEM error will occur if the driver constant ETTPOOL_SZ
 *	is too small. EINVAL error occurs for a variety of situations:
 *	- driver not inited
 *	- some te_len is < TG_MINDPL or > TG_MAXDPL or some te_Dt is null
 *	- the last entry in the timetable is not null
 *	- the used packet length crosses a click boundary
 *	- the code of the message is not less than CTL_CODE.
 * TG_KILL
 *	Completely erases a TGT. The controller must supply the address of the
 *	TG data packet. The space for the TG data packet can be disposed
 *	after this command has been issued. The driver supplies the
 *	sequence number of the last packet sent by the killed TGT.
 * TG_RATE
 *	Changes the throughput of a TGT by multiplying the Dt or the length's
 *	of its timetable by two given factors. Factors equal to 1 bring
 *	the throughput back to the original.
 * TG_ON
 *	The activity flag of the addressed TGT is set to ON. This means
 *	that the TGT immediately begins to transmit messages to the FODA
 *	system. If the activity is already ON, the command is ignored.
 * TG_OFF
 *	The activity flag of the addressed TGT is set to OFF. This means
 *	that the TGT immediately stops transmitting messages to the FODA
 *	system. If the activity is already OFF, the command is ignored.
 */

/* union used for ioctl's	*/
typedef union {
/************* contiguous receive buffer control ************************/
	struct {
		caddr_t	physaddr;
		uint	size;
	} crb;
/************* MTG initialization ***************************************/
	struct {
		/* input data (controller -> driver)	*/
		caddr_t	ucrbuf;		/* controller address of crbuf	*/
		cmp_e	*ucmp_a;	/* controller address of CMP array */
		dpp_e	*udpp_a;	/* controller address of DPP array */
		uint	max_sem;	/* max semaphore value allowed	*/
		time_t	tick_len;	/* length of tick in usec	*/
		uint	ettpool_sz;	/* size of ett_e's pool		*/
		/* output data (driver -> controller)	*/
		time_t	*utick;		/* controller address of microtick */
		int	*sema;		/* controller address of semaphore */
	} init;
/************* CTL message **********************************************/
	cm_t	*ucm;
/************* TGT creation *********************************************/
	struct {
		dp_t	*udp;		/* controller address of data packet */
		tt_e	*utt;		/* u.a. of 1st timetable entry	*/
		uint	ttsz;		/* # of slots in the timetable	*/
	} create;
/************* TGT deletion *********************************************/
	struct {
		dp_t	*udp;		/* controller address of data packet */
		ushort	last_seq;	/* seq.no. of last packet sent	*/
	} kill;
/************* TGT throughput change ************************************/
	struct {
		dp_t	*udp;		/* controller address of data packet */
		float	Dt_f,		/* factor to multiply te_Dt's	*/
			len_f;		/* factor to multiply te_len's	*/
	} rate;
/***** controller address of data packet used for on and off *****************/
	dp_t	*udp;
} mtg_ctl;


/* Commands for ioctl	*/
#define	TG_IOCTL_PREFIX	(unchar)'M'<<3*8 | (unchar)'T'<<2*8 | (unchar)'G'<<1*8
#define	TG_GETRBUF	(TG_IOCTL_PREFIX | 0x01)
#define	TG_FREERBUF	(TG_IOCTL_PREFIX | 0x02)
#define	TG_ASKRBUF	(TG_IOCTL_PREFIX | 0x03)
#define	TG_RESET	(TG_IOCTL_PREFIX | 0x11)
#define	TG_INIT		(TG_IOCTL_PREFIX | 0x20)
#define	TG_HALT		(TG_IOCTL_PREFIX | 0x21)
#define	TG_CTL		(TG_IOCTL_PREFIX | 0x22)
#define	TG_CREATE	(TG_IOCTL_PREFIX | 0x23)
#define	TG_KILL		(TG_IOCTL_PREFIX | 0x24)
#define	TG_RATE		(TG_IOCTL_PREFIX | 0x25)
#define	TG_ON		(TG_IOCTL_PREFIX | 0x26)
#define	TG_OFF		(TG_IOCTL_PREFIX | 0x27)

/****************************************************************/
/*			 MTG error types			*/
/****************************************************************/

/* all error messages contain a timestamp in cm_MTG_h.mtx_time	*/
#define	TE_SLOW	0x01
#define	TE_LCOL	0x02
#define	TE_LCAR 0x03
#define	TE_RTRY	0x04
#define	TE_SEMA	0x05
/*   error TE_SLOW is fatal from the controller's point of view:	*/
/* any ioctl from there on (apart from TG_HALT and the crbuf commands)	*/
/* generates an 'invalid argument' error.				*/

/* this macro can be used to initialize an array of error messages	*/
#define	ERRINIT	{							  \
	"Error on error message (should never happen) !!!!!",	\
	"Receive buffer overflow",		/* TE_SLOW */	\
	"Late collision on transmission",	/* TE_LCOL */	\
	"Loss of carrier on transmission",	/* TE_LCAR */	\
	"Retry error on transmission",		/* TE_RTRY */	\
	"Too much backlog",			/* TE_SEMA */	\
	(char *) NULL						\
	}

/****************************************************************/
/*	Miscellaneous constants inside the driver		*/
/****************************************************************/

#define	TICK_RATIO	10	/* (tick length)/(microtick length)	*/
#define ETTPOOL_MIN_SZ	100	/* min. for mtg_ctl.init.ettpool_sz	*/
